# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Representations of boolean expressions and variables.

These classes can be used to represent simple boolean
expressions and do evaluation with substitutions of
particular variables.
"""

import re


class Variable(object):
    """Represents a variable in a boolean expression"""

    def __init__(self, symbol):
        self._symbol = str(symbol)

    @property
    def symbol(self):
        return self._symbol

    def substitute(self, mapping):
        """Substitute variables using mapping function"""
        return mapping(self)

    def __repr__(self):
        return 'Variable({})'.format(repr(self._symbol))

    def __str__(self):
        return str(self._symbol)

    def __eq__(self, other):
        return isinstance(other, Variable) and self._symbol == other._symbol

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash('Variable') ^ hash(self._symbol)


class And(object):
    """Represents a conjuction of boolean terms"""

    def __init__(self, *args):
        terms = set()
        for t in args:
            if isinstance(t, And):
                terms.update(t.terms)
            else:
                terms.add(t)
        self._terms = frozenset(terms)

    @property
    def terms(self):
        """Iterate over terms"""
        return iter(self._terms)

    def substitute(self, mapping):
        """Substitute variables using mapping function"""
        result = []
        for t in self._terms:
            value = t.substitute(mapping)
            if isinstance(value, bool):
                if not value:
                    return False
            else:
                result.append(value)
        if len(result) == 0:
            return True
        elif len(result) == 1:
            return result[0]
        return And(*result)

    def __repr__(self):
        return 'Expression({})'.format(repr(str(self)))

    def __str__(self):
        def format_term(t):
            if isinstance(t, Variable):
                return str(t)
            return '({})'.format(t)
        return ' and '.join(format_term(t) for t in self._terms)

    def __eq__(self, other):
        if isinstance(other, Variable) and len(self._terms) == 1:
            return self._terms[0] == other
        return isinstance(other, And) and self._terms == other._terms

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash('And') ^ hash(self._terms)


class Or(object):
    """Represents a disjuction of boolean terms"""

    def __init__(self, *args):
        terms = set()
        for t in args:
            if isinstance(t, Or):
                terms.update(t.terms)
            else:
                terms.add(t)
        self._terms = frozenset(terms)

    @property
    def terms(self):
        """Iterate over terms"""
        return iter(self._terms)

    def substitute(self, mapping):
        """Substitute variables using mapping function"""
        result = []
        for t in self._terms:
            value = t.substitute(mapping)
            if isinstance(value, bool):
                if value:
                    return True
            else:
                result.append(value)
        if len(result) == 0:
            return False
        elif len(result) == 1:
            return result[0]
        return Or(*result)

    def __repr__(self):
        return 'Expression({})'.format(repr(str(self)))

    def __str__(self):
        def format_term(t):
            if isinstance(t, Variable):
                return str(t)
            return '({})'.format(t)
        return ' or '.join(format_term(t) for t in self._terms)

    def __eq__(self, other):
        if isinstance(other, Variable) and len(self._terms) == 1:
            return self._terms[0] == other
        return isinstance(other, Or) and self._terms == other._terms

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash('Or') ^ hash(self._terms)


def Expression(s):  # noqa
    """Parse boolean expression containing and/or operators"""

    # Converters for opeartor clauses
    operators = {
        'and': And,
        'or': Or,
        None: lambda *args: args[0]
    }

    scanner = re.compile(r'''
        (\s+) |              # space
        (\(|\)) |            # group
        ((?:or|and)\b) |     # operator
        ([^\d\W]\w*) |       # variable
        (\Z) |               # end
        (.)                  # error
    ''', re.DOTALL | re.VERBOSE)

    # Parsed using two states and a stack of open clauses
    # At state 0 (not expect_operator): Expect variable, or parenthesis group
    #  start.
    # At state 1 (expect_operator): Expect operator, parenthesis group end, or
    #  end.
    expect_operator = False
    clause_stack = []
    current_clause = []
    clause_operator = None

    def close():
        prev_op, prev_clause = clause_stack.pop()
        prev_clause.append(operators[clause_operator](*current_clause))
        return prev_op, prev_clause

    for match in re.finditer(scanner, s):
        space, group, operator, variable, end, error = match.groups()

        if error is not None:
            raise ValueError('Invalid token in expression string: {}'.format(
                repr(match.group(0))))
        elif space is not None:
            continue
        elif expect_operator and operator is not None:
            if operator == 'and' and clause_operator != 'and':
                prev_term = current_clause.pop()
                clause_stack.append((clause_operator, current_clause))
                current_clause = [prev_term]
            elif operator == 'or' and clause_operator == 'and':
                clause_operator, current_clause = close()
            clause_operator = operator
            expect_operator = False
        elif expect_operator and group is not None and group == ')':
            if clause_operator == 'and':
                clause_operator, current_clause = close()
            if len(clause_stack) == 0:
                raise ValueError('Unbalanced parenthesis group in expression')
            clause_operator, current_clause = close()
        elif expect_operator and end is not None:
            if clause_operator == 'and':
                clause_operator, current_clause = close()
        elif not expect_operator and variable is not None:
            current_clause.append(Variable(variable))
            expect_operator = True
        elif not expect_operator and group is not None and group == '(':
            clause_stack.append((clause_operator, current_clause))
            current_clause = []
            clause_operator = None
        else:
            raise ValueError('Invalid token in expression string: {}'.format(
                repr(match.group(0))))

    if len(clause_stack) > 0:
        raise ValueError('Unbalanced parenthesis group in expression')

    expr = operators[clause_operator](*current_clause)
    return expr



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.formula import Formula, Atom, Radical


class TestFormula(unittest.TestCase):
    def test_formula_merge_same_formulas_with_same_atoms(self):
        f = Formula({Atom('H'): 2, Atom('O'): 1}) | Formula({Atom('N'): 1, Atom('O'): 2})
        self.assertEqual(f, Formula({Atom('H'): 2, Atom('N'): 1, Atom('O'): 3}))

    def test_formula_merge_formulas_that_cancel_out(self):
        f = Formula({Atom('H'): 3}) | Formula({Atom('H'): -3})
        self.assertEqual(f, Formula())

    def test_formula_multiply_number(self):
        f = Formula({Atom('H'): 2, Atom('O'): 1}) * 4
        self.assertEqual(f, Formula({Atom('H'): 8, Atom('O'): 4}))

    def test_formula_multiply_one(self):
        f = Formula({Atom('H'): 2, Atom('O'): 1}) * 1
        self.assertEqual(f, f)

    def test_formula_multiply_zero(self):
        f = Formula({Atom('H'): 2, Atom('O'): 1}) * 0
        self.assertEqual(f, Formula())

    def test_formula_right_multiply_number(self):
        f = 2 * Formula({Atom('H'): 2, Atom('O'): 1})
        self.assertEqual(f, Formula({Atom('H'): 4, Atom('O'): 2}))

    def test_formula_repeat(self):
        f = Formula({Atom('H'): 2, Atom('O'): 1}).repeat(4)
        self.assertEqual(f, Formula({ Formula({Atom('H'): 2, Atom('O'): 1}): 4 }))

    def test_formula_equals_other_formula(self):
        f = Formula({Atom('H'): 2, Atom('O'): 1})
        self.assertEqual(f, Formula({Atom('O'): 1, Atom('H'): 2}))

    def test_formula_not_equals_other_with_distinct_elements(self):
        f = Formula({Atom('Au'): 1})
        self.assertNotEqual(f, Formula({Atom('Ag'): 1}))

    def test_formula_not_equals_other_with_different_number(self):
        f = Formula({Atom('Au'): 1})
        self.assertNotEqual(f, Formula({Atom('Au'): 2}))

    def test_formula_parse_with_final_digit(self):
        f = Formula.parse('H2O2')
        self.assertEqual(f, Formula({Atom('H'): 2, Atom('O'): 2}))

    def test_formula_parse_with_implicit_final_digit(self):
        f = Formula.parse('H2O')
        self.assertEqual(f, Formula({Atom('H'): 2, Atom('O'): 1}))

    def test_formula_parse_with_implicit_digit(self):
        f = Formula.parse('C2H5NO2')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('H'): 5, Atom('N'): 1, Atom('O'): 2}))

    def test_formula_parse_with_wide_element(self):
        f = Formula.parse('ZnO')
        self.assertEqual(f, Formula({Atom('Zn'): 1, Atom('O'): 1}))

    def test_formula_parse_with_wide_count(self):
        f = Formula.parse('C6H10O2')
        self.assertEqual(f, Formula({Atom('C'): 6, Atom('H'): 10, Atom('O'): 2}))

    def test_formula_parse_with_implicitly_counted_subgroup(self):
        f = Formula.parse('C2H6O2(CH)')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('H'): 6, Atom('O'): 2,
                                      Formula({Atom('C'): 1, Atom('H'): 1}): 1}))

    def test_formula_parse_with_counted_subgroup(self):
        f = Formula.parse('C2H6O2(CH)2')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('H'): 6, Atom('O'): 2,
                                      Formula({Atom('C'): 1, Atom('H'): 1}): 2}))

    def test_formula_parse_with_two_identical_counted_subgroups(self):
        f = Formula.parse('C2H6O2(CH)2(CH)2')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('H'): 6, Atom('O'): 2,
                                      Formula({Atom('C'): 1, Atom('H'): 1}): 4}))

    def test_formula_parse_with_two_distinct_counted_subgroups(self):
        f = Formula.parse('C2H6O2(CH)2(CH2)2')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('H'): 6, Atom('O'): 2,
                                      Formula({Atom('C'): 1, Atom('H'): 1}): 2,
                                      Formula({Atom('C'): 1, Atom('H'): 2}): 2}))

    def test_formula_parse_with_wide_counted_subgroup(self):
        f = Formula.parse('C2(CH)10NO2')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('N'): 1, Atom('O'): 2,
                                        Formula({Atom('C'): 1, Atom('H'): 1}): 10}))

    def test_formula_parse_with_radical(self):
        f = Formula.parse('C2H4NO2R')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('H'): 4, Atom('N'): 1,
                                        Atom('O'): 2, Radical('R'): 1}))

    def test_formula_parse_with_numbered_radical(self):
        f = Formula.parse('C2H4NO2(R1)')
        self.assertEqual(f, Formula({Atom('C'): 2, Atom('H'): 4, Atom('N'): 1,
                                        Atom('O'): 2, Radical('R1'): 1}))

    def test_formula_is_not_variable(self):
        f = Formula.parse('C6H12O6')
        self.assertFalse(f.is_variable())

    def test_formula_is_variable(self):
        f = Formula.parse('C2H4NO2R(C2H2NOR)n')
        self.assertTrue(f.is_variable())

    def test_formula_balance_missing_on_one_side(self):
        f1, f2 = Formula.balance(Formula.parse('H2O'), Formula.parse('OH'))
        self.assertEqual(f1, Formula())
        self.assertEqual(f2, Formula({Atom('H'): 1}))

    def test_formula_balance_missing_on_both_sides(self):
        f1, f2 = Formula.balance(Formula.parse('C3H6OH'), Formula.parse('CH6O2'))
        self.assertEqual(f1, Formula({Atom('O'): 1}))
        self.assertEqual(f2, Formula({Atom('C'): 2, Atom('H'): 1}))

    def test_formula_balance_subgroups_cancel_out(self):
        f1, f2 = Formula.balance(Formula.parse('H2(CH2)n'), Formula.parse('CH3O(CH2)n'))
        self.assertEqual(f1, Formula({Atom('C'): 1, Atom('H'): 1, Atom('O'): 1}))
        self.assertEqual(f2, Formula())


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Fastcore module implementing the fastcore algorithm

This is an implementation of the algorithms described in [Vlassis14]_.
"""

import logging

from .lpsolver import lp
from .fluxanalysis import flux_balance
from .metabolicmodel import FlipableModelView

from six import iteritems

# Module-level logging
logger = logging.getLogger(__name__)


def support(fluxiter, threshold=None):
    """Yield reactions in the support set of the fluxes

    The fluxiter argument must be an iterable returning reaction, flux-pairs.
    The support set is the set of reactions that have an absolute flux value at
    or above a certain threshold. If threshold is None the mathematical
    definition (v != 0) is used.
    """

    if threshold is None:
        return (rxnid for rxnid, v in fluxiter if v != 0)
    return (rxnid for rxnid, v in fluxiter if abs(v) >= threshold)


def support_positive(fluxiter, threshold=None):
    """Yield reactions in the support set of the fluxes (only positive fluxes)

    The fluxiter argument must be an iterable returning reaction, flux-pairs.
    The support set is the set of reactions that have a flux value at or above
    a certain threshold. If threshold is None the mathematical definition
    (v > 0) is used.
    """

    if threshold is None:
        return (rxnid for rxnid, v in fluxiter if v > 0)
    return (rxnid for rxnid, v in fluxiter if v >= threshold)


class FastcoreError(Exception):
    """Indicates an error while running Fastcore"""


def lp7(model, reaction_subset, epsilon, solver):
    """Approximately maximize the number of reaction with flux above epsilon

    This is similar to FBA but approximately maximizing the number of reactions
    in subset with flux > epsilon, instead of just maximizing the flux of one
    particular reaction. LP7 prefers "flux splitting" over
    "flux concentrating".
    """

    # Create LP-7 problem of Fastcore
    prob = solver.create_problem()

    # Define flux variables
    for rxnid in model.reactions:
        lower, upper = model.limits[rxnid]
        prob.define(('v', rxnid), lower=lower, upper=upper)

    # Define z variables
    prob.define(*(('z', rxnid) for rxnid in reaction_subset),
                lower=0, upper=epsilon)
    prob.set_linear_objective(prob.expr(
        {('z', rxnid): 1 for rxnid in reaction_subset}))
    v = prob.set(('v', rxnid) for rxnid in reaction_subset)
    z = prob.set(('z', rxnid) for rxnid in reaction_subset)
    prob.add_linear_constraints(v >= z)

    massbalance_lhs = {compound: 0 for compound in model.compounds}
    for spec, value in iteritems(model.matrix):
        compound, rxnid = spec
        massbalance_lhs[compound] += prob.var(('v', rxnid)) * value
    prob.add_linear_constraints(
        *(lhs == 0 for compound, lhs in iteritems(massbalance_lhs)))

    # Solve
    result = prob.solve(lp.ObjectiveSense.Maximize)
    if not result:
        raise FastcoreError('Non-optimal solution: {}'.format(result.status))

    for rxnid in sorted(model.reactions):
        yield rxnid, result.get_value(('v', rxnid))


def lp10(model, subset_k, subset_p, epsilon, scaling, solver, weights={}):
    """Force reactions in K above epsilon while minimizing support of P

    This program forces reactions in subset K to attain flux > epsilon
    while minimizing the sum of absolute flux values for reactions
    in subset P (L1-regularization).
    """

    if len(subset_k) == 0:
        return

    # Create LP-10 problem of Fastcore
    prob = solver.create_problem()

    # Define flux variables
    for rxnid in model.reactions:
        lower, upper = model.limits[rxnid]
        if rxnid in subset_k:
            lower = max(lower, epsilon)
        prob.define(('v', rxnid), lower=lower*scaling, upper=upper*scaling)

    # Define z variables
    prob.define(*(('z', rxnid) for rxnid in subset_p), lower=0)
    prob.set_linear_objective(prob.expr(
        {('z', rxnid): weights.get(rxnid, 1) for rxnid in subset_p}))

    z = prob.set(('z', rxnid) for rxnid in subset_p)
    v = prob.set(('v', rxnid) for rxnid in subset_p)
    prob.add_linear_constraints(z >= v, v >= -z)

    massbalance_lhs = {compound: 0 for compound in model.compounds}
    for spec, value in iteritems(model.matrix):
        compound, rxnid = spec
        massbalance_lhs[compound] += prob.var(('v', rxnid)) * value
    prob.add_linear_constraints(
        *(lhs == 0 for compound, lhs in iteritems(massbalance_lhs)))

    # Solve
    result = prob.solve(lp.ObjectiveSense.Minimize)
    if not result:
        raise FastcoreError('Non-optimal solution: {}'.format(result.status))

    for reaction_id in model.reactions:
        yield reaction_id, result.get_value(('v', reaction_id))


def fastcc(model, epsilon, solver):
    """Check consistency of model reactions

    Yields all reactions in the model that are not part
    of the consistent subset.
    """

    reaction_set = set(model.reactions)
    subset = reaction_set.difference(model.reversible)

    logger.debug('|J| = {}, J = {}'.format(len(subset), subset))

    consistent_subset = set(
        support(lp7(model, subset, epsilon, solver), epsilon))

    logger.debug('|A| = {}, A = {}'.format(
        len(consistent_subset), consistent_subset))

    for reaction in subset - consistent_subset:
        # Inconsistent reaction
        yield reaction

    # Check remaining reactions
    subset = (reaction_set - subset) - consistent_subset

    logger.debug('|J| = {}, J = {}'.format(len(subset), subset))

    # Wrap model in flipable proxy so reactions can be flipped
    model = FlipableModelView(model)

    flipped = False
    singleton = False
    while len(subset) > 0:
        if singleton:
            reaction = next(iter(subset))
            subset_i = {reaction}

            logger.debug('LP3 on {}'.format(subset_i))
            supp = support(flux_balance(
                model, reaction, tfba=False, solver=solver), epsilon)
        else:
            subset_i = subset

            logger.debug('LP7 on {}'.format(subset_i))
            supp = support(lp7(model, subset_i, epsilon, solver), epsilon)
        consistent_subset.update(supp)

        logger.debug('|A| = {}, A = {}'.format(
            len(consistent_subset), consistent_subset))

        if not subset.isdisjoint(consistent_subset):
            subset -= consistent_subset
            logger.debug('|J| = {}, J = {}'.format(len(subset), subset))
            flipped = False
        else:
            # TODO: irreversible reactions are taken care of before the
            # loop so at this point all reactions in subset_i are reversble(?).
            subset_rev_i = subset_i & model.reversible
            if flipped or len(subset_rev_i) == 0:
                flipped = False
                if singleton:
                    subset -= subset_rev_i
                    for reaction in subset_rev_i:
                        yield reaction
                else:
                    singleton = True
            else:
                model.flip(subset_rev_i)
                flipped = True
                logger.debug('Flip')


def fastcc_is_consistent(model, epsilon, solver):
    """Quickly check whether model is consistent

    Returns true if the model is consistent. If it is only necessary to know
    whether a model is consistent, this function is faster as it will return
    the result as soon as it finds a single inconsistent reaction.
    """

    for reaction in fastcc(model, epsilon, solver):
        return False
    return True


def fastcc_consistent_subset(model, epsilon, solver):
    """Return consistent subset of model

    The largest consistent subset is returned as
    a set of reaction names.
    """

    reaction_set = set(model.reactions)
    return reaction_set.difference(fastcc(model, epsilon, solver))


def find_sparse_mode(model, core, additional, epsilon, scaling, solver,
                     weights={}):
    """Find a sparse mode containing reactions of the core subset

    Return an iterator of the support of a sparse mode that contains as many
    reactions from core as possible, and as few reactions from additional as
    possible (approximately). A dictionary of weights can be supplied which
    gives further penalties for including specific additional reactions.
    """

    if len(core) == 0:
        return iter(())

    supp = support_positive(lp7(model, core, epsilon, solver), epsilon)
    k = core.intersection(supp)
    if len(k) == 0:
        return iter(())

    return support(lp10(model, k, additional, epsilon, solver=solver,
                        scaling=scaling, weights=weights), epsilon)


def fastcore(model, core, epsilon, solver, scaling=1e8, weights={}):
    """Find a flux consistent subnetwork containing the core subset

    The result will contain the core subset and as few of the additional
    reactions as possible.
    """

    consistent_subset = set()
    reaction_set = set(model.reactions)

    subset = core - model.reversible
    logger.debug('|J| = {}, J = {}'.format(len(subset), subset))

    penalty_set = reaction_set - core
    logger.debug('|P| = {}, P = {}'.format(len(penalty_set), penalty_set))

    mode = set(find_sparse_mode(
        model, subset, penalty_set, epsilon, solver=solver, scaling=scaling,
        weights=weights))
    if not subset.issubset(mode):
        raise FastcoreError('Inconsistent irreversible core reactions:'
                            ' {}'.format(subset - mode))

    consistent_subset |= mode
    logger.debug('|A| = {}, A = {}'.format(
        len(consistent_subset), consistent_subset))

    subset = core - mode
    logger.debug('|J| = {}, J = {}'.format(len(subset), subset))

    # Wrap model in flipable proxy so reactions can be flipped
    model = FlipableModelView(model)

    flipped = False
    singleton = False
    while len(subset) > 0:
        penalty_set -= consistent_subset
        if singleton:
            subset_i = set((next(iter(subset)),))
        else:
            subset_i = subset

        mode = find_sparse_mode(
            model, subset_i, penalty_set, epsilon, scaling=scaling,
            solver=solver, weights=weights)
        consistent_subset.update(mode)
        logger.debug('|A| = {}, A = {}'.format(
            len(consistent_subset), consistent_subset))

        if not subset.isdisjoint(consistent_subset):
            logger.debug('Subset improved {} -> {}'.format(
                len(subset), len(subset - consistent_subset)))
            subset -= consistent_subset
            logger.debug('|J| = {}, J = {}'.format(len(subset), subset))
            flipped = False
        else:
            logger.debug('Nothing found, changing state...')
            subset_rev_i = subset_i & model.reversible
            if flipped or len(subset_rev_i) == 0:
                if singleton:
                    raise FastcoreError('Global network inconsistent:'
                                        ' {}'.format(subset_rev_i))

                logger.debug('Going to non-flipped, singleton state...')
                singleton = True
                flipped = False
            else:
                model.flip(subset_rev_i)
                flipped = True
                logger.debug('Going to flipped state... {}'.format(
                    subset_rev_i))

    return consistent_subset



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Parser for SBML model files"""

import xml.etree.ElementTree as ET
from decimal import Decimal
from fractions import Fraction
from functools import partial
from itertools import count
import logging

from six import itervalues, iteritems

from ..reaction import Reaction, Compound


logger = logging.getLogger(__name__)


# Level 1 namespaces
SBML_NS_L1 = 'http://www.sbml.org/sbml/level1'

# Level 2 namespaces
SBML_NS_L2 = 'http://www.sbml.org/sbml/level2'
SBML_NS_L2_V2 = 'http://www.sbml.org/sbml/level2/version2'
SBML_NS_L2_V3 = 'http://www.sbml.org/sbml/level2/version3'
SBML_NS_L2_V4 = 'http://www.sbml.org/sbml/level2/version4'
SBML_NS_L2_V5 = 'http://www.sbml.org/sbml/level2/version5'

# Level 3 namespaces
SBML_NS_L3_V1_CORE = 'http://www.sbml.org/sbml/level3/version1/core'

MATHML_NS = 'http://www.w3.org/1998/Math/MathML'


def _tag(tag, namespace=None):
    """Prepend namespace to tag name"""
    if namespace is None:
        return str(tag)
    return '{{{}}}{}'.format(namespace, tag)


class ParseError(Exception):
    """Error parsing SBML file"""


class _SBMLEntry(object):
    """Base class for compound and reaction entries"""

    def __init__(self, reader, root):
        self._reader = reader
        self._root = root
        self._id = self._element_get_id(root)

    def _element_get_id(self, element):
        """Get id of reaction or species element.

        In old levels the name is used as the id. This method returns the
        correct attribute depending on the level.
        """
        if self._reader._level > 1:
            entry_id = element.get('id')
        else:
            entry_id = element.get('name')
        return entry_id

    @property
    def id(self):
        """Entity ID"""
        return self._id

    @property
    def xml_notes(self):
        """Access the entity notes as an XML document fragment"""
        return self._root.find(self._reader._sbml_tag('notes'))


class SpeciesEntry(_SBMLEntry):
    """Species entry in the SBML file"""

    def __init__(self, reader, root):
        super(SpeciesEntry, self).__init__(reader, root)

        self._name = root.get('name')
        self._comp = root.get('compartment')

        self._boundary = root.get('boundaryCondition', 'false') == 'true'

        # In non-strict mode the species that ends with _b are considered
        # boundary conditions.
        if not self._reader._strict and self._id.endswith('_b'):
            logger.warning('Species {} was converted to boundary condition'
                           ' because of "_b" suffix'.format(self.id))
            self._boundary = True

    @property
    def name(self):
        """Species name"""
        return self._name

    @property
    def compartment(self):
        """Species compartment"""
        return self._comp

    @property
    def charge(self):
        """Species charge"""
        charge = self._root.get('charge')
        if charge is not None and charge != '':
            return int(charge)
        return None

    @property
    def boundary(self):
        """Whether this compound is a boundary condition"""
        return self._boundary

    @property
    def properties(self):
        """All species properties as a dict"""
        properties = {'id': self._id,
                      'boundary': self._boundary}
        if 'name' in self._root.attrib:
            properties['name'] = self._root.get('name')
        if 'compartment' in self._root.attrib:
            properties['compartment'] = self._root.get('compartment')
        if 'charge' in self._root.attrib and self._root.get('charge') != '':
            properties['charge'] = int(self._root.get('charge'))

        return properties


class ReactionEntry(_SBMLEntry):
    """Reaction entry in SBML file"""

    def __init__(self, reader, root):
        super(ReactionEntry, self).__init__(reader, root)

        self._name = self._root.get('name')
        self._rev = self._root.get('reversible', 'true') == 'true'

        left, right = [], []
        for side, tag_name in ((left, 'listOfReactants'),
                               (right, 'listOfProducts')):
            for species_id, value in self._parse_species_references(tag_name):
                try:
                    species_entry = self._reader.get_species(species_id)
                    if (self._reader._ignore_boundary and
                            species_entry.boundary):
                        # Skip boundary species when ignoring these
                        continue
                except KeyError:
                    message = ('Reaction {} references non-existent'
                               ' species {}'.format(self._id, species_id))
                    if not self._reader._strict:
                        # In non-strict mode simply skip these references
                        logger.warn(message)
                        continue
                    raise ParseError(message)

                species_id, species_comp = (
                    species_entry.id, species_entry.compartment)
                compound = Compound(species_id, compartment=species_comp)
                side.append((compound, value))

        # Add reaction to database
        direction = Reaction.Bidir if self._rev else Reaction.Right
        self._equation = Reaction(direction, left, right)

    def _parse_species_references(self, name):
        """Yield species id and parsed value for a speciesReference list"""
        for species in self._root.iterfind('./{}/{}'.format(
                self._reader._sbml_tag(name),
                self._reader._sbml_tag('speciesReference'))):

            species_id = species.get('species')

            if self._reader._level == 1:
                # In SBML level 1 only positive integers are allowed for
                # stoichiometry but a positive integer denominator can be
                # specified.
                try:
                    value = int(species.get('stoichiometry', 1))
                    denom = int(species.get('denominator', 1))
                    species_value = Fraction(value, denom)
                except ValueError:
                    message = ('Non-integer stoichiometry is not allowed in'
                               ' SBML level 1 (reaction {})'.format(self.id))
                    if self._reader._strict:
                        raise ParseError(message)
                    else:
                        logger.warning(message)
                    species_value = Decimal(species.get('stoichiometry', 1))
            elif self._reader._level == 2:
                # Stoichiometric value is a double but can alternatively be
                # specified using math (not implemented).
                value_str = species.get('stoichiometry', None)
                if value_str is None:
                    if 'stoichiometryMath' in species:
                        raise ParseError('stoichiometryMath in '
                                         'speciesReference is not implemented')
                    species_value = 1
                else:
                    species_value = Decimal(value_str)
            elif self._reader._level == 3:
                # Stoichiometric value is a double but can alternatively be
                # specified using initial assignment (not implemented).
                value_str = species.get('stoichiometry', None)
                if value_str is None:
                    raise ParseError('Missing stoichiometry in'
                                     ' speciesReference is not implemented')
                species_value = Decimal(value_str)

            if species_value % 1 == 0:
                species_value = int(species_value)

            yield species_id, species_value

    @property
    def id(self):
        """Reaction ID"""
        return self._id

    @property
    def name(self):
        """Reaction name"""
        return self._name

    @property
    def reversible(self):
        """Whether the reaction is reversible"""
        return self._rev

    @property
    def equation(self):
        """Reaction equation is a :class:`Reaction <psamm.reaction.Reaction>`
        object"""
        return self._equation

    @property
    def kinetic_law_reaction_parameters(self):
        """Iterator over the values of kinetic law reaction parameters"""

        for parameter in self._root.iterfind(
                './{}/{}/{}'.format(self._reader._sbml_tag('kineticLaw'),
                                    self._reader._sbml_tag('listOfParameters'),
                                    self._reader._sbml_tag('parameter'))):
            param_id = parameter.get('id')
            param_name = parameter.get('name')
            param_value = float(parameter.get('value'))
            param_units = parameter.get('units')

            yield param_id, param_name, param_value, param_units

    @property
    def properties(self):
        """All reaction properties as a dict"""
        properties = {'id': self._id,
                      'reversible': self._rev,
                      'equation': self._equation}
        if 'name' in self._root.attrib:
            properties['name'] = self._root.get('name')

        return properties


class SBMLReader(object):
    """Reader of SBML model files

    The constructor takes a file-like object which will be parsed as XML and
    then as SBML according to the specification. If the ``strict`` parameter is
    set to False, the parser will revert to a more lenient parsing which is
    required for many older models. This tries to mimic the inconsistencies
    employed by COBRA when parsing models.

    If ``ignore_boundary`` is ``True``, the species that are marked as
    boundary conditions will simply be dropped from the species list and from
    the reaction equations.
    """

    def __init__(self, file, strict=False, ignore_boundary=False):
        # Parse SBML file
        tree = ET.parse(file)
        root = tree.getroot()

        self._strict = strict
        self._ignore_boundary = ignore_boundary

        # Parse level and version
        self._sbml_tag = None
        self._level = int(root.get('level'))
        self._version = int(root.get('version'))

        if self._level == 1:
            self._sbml_tag = partial(_tag, namespace=SBML_NS_L1)
        elif self._level == 2:
            if self._version == 1:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2)
            elif self._version == 2:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V2)
            elif self._version == 3:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V3)
            elif self._version == 4:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V4)
            elif self._version == 5:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V5)
            else:
                if self._strict:
                    raise ParseError('SBML level 2, version {}'
                                     ' not implemented'.format(self._version))
                # Assume latest version
                self._version = 5
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V5)
        elif self._level == 3:
            if self._version == 1:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L3_V1_CORE)
            else:
                if self._strict:
                    raise ParseError('SBML level 3, version {}'
                                     ' not implemented'.format(self._version))
                # Assume latest version
                self._version = 1
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L3_V1_CORE)
        else:
            if self._strict:
                raise ParseError('SBML level {} not implemented'.format(
                    self._level))
            # Assume level 1
            self._level = 1
            self._sbml_tag = partial(_tag, namespace=SBML_NS_L1)

        self._model = root.find(self._sbml_tag('model'))

        # Species
        self._model_species = {}
        self._species = self._model.find(self._sbml_tag('listOfSpecies'))
        for species in self._species.iterfind(self._sbml_tag('species')):
            entry = SpeciesEntry(self, species)
            self._model_species[entry.id] = entry

        # Reactions
        self._model_reactions = {}
        self._reactions = self._model.find(self._sbml_tag('listOfReactions'))
        for reaction in self._reactions.iterfind(self._sbml_tag('reaction')):
            entry = ReactionEntry(self, reaction)
            self._model_reactions[entry.id] = entry

    def get_reaction(self, reaction_id):
        """Return :class:`.ReactionEntry` corresponding to reaction_id"""
        return self._model_reactions[reaction_id]

    def get_species(self, species_id):
        """Return :class:`.SpeciesEntry` corresponding to species_id"""
        return self._model_species[species_id]

    @property
    def reactions(self):
        """Iterator over :class:`ReactionEntries <.ReactionEntry>`"""
        return itervalues(self._model_reactions)

    @property
    def species(self):
        """Iterator over :class:`SpeciesEntries <.SpeciesEntry>`

        This will not yield boundary condition species if those are ignored.
        """
        return (c for c in itervalues(self._model_species)
                if not self._ignore_boundary or not c.boundary)

    @property
    def id(self):
        """Model ID"""
        return self._model.get('id', None)

    @property
    def name(self):
        """Model name"""
        return self._model.get('name', None)


class SBMLWriter(object):
    """Writer of SBML files"""

    def __init__(self):
        self._namespace = SBML_NS_L3_V1_CORE
        self._sbml_tag = partial(_tag, namespace=self._namespace)

    def write_model(self, file, model, compounds):
        """Write a given model to file"""

        ET.register_namespace('mathml', MATHML_NS)

        # Load compound information
        compound_name = {}
        for compound in compounds:
            compound_name[compound.id] = (
                compound.name if compound.name is not None else compound.id)

        root = ET.Element(self._sbml_tag('sbml'))
        root.set(self._sbml_tag('level'), '3')
        root.set(self._sbml_tag('version'), '1')

        model_tag = ET.SubElement(root, self._sbml_tag('model'))

        # Generators of unique IDs
        compound_id = ('M_'+str(i) for i in count(1))
        compartment_id = ('C_'+str(i) for i in count(1))
        reaction_id = ('R_'+str(i) for i in count(1))

        # Build mapping from Compound to species ID
        model_compartments = {}
        model_species = {}
        for reaction in model.reactions:
            for compound, value in model.get_reaction_values(reaction):
                if compound.compartment not in model_compartments:
                    model_compartments[compound.compartment] = (
                        next(compartment_id))
                if compound not in model_species:
                    model_species[compound] = next(compound_id)

        # Create list of compartments
        compartments = ET.SubElement(
            model_tag, self._sbml_tag('listOfCompartments'))
        for compartment, compartment_id in iteritems(model_compartments):
            compartment_tag = ET.SubElement(
                compartments, self._sbml_tag('compartment'))
            compartment_tag.set(self._sbml_tag('id'), compartment_id)
            compartment_tag.set(self._sbml_tag('name'), str(compartment))

        # Create list of species
        species_list = ET.SubElement(
            model_tag, self._sbml_tag('listOfSpecies'))
        for species, species_id in iteritems(model_species):
            species_tag = ET.SubElement(species_list,
                                        self._sbml_tag('species'))
            species_tag.set(self._sbml_tag('id'), species_id)
            species_tag.set(
                self._sbml_tag('name'),
                str(species.translate(lambda x: compound_name.get(x, x))))
            species_tag.set(
                self._sbml_tag('compartment'),
                model_compartments[species.compartment])

        # Create list of reactions
        reactions = ET.SubElement(model_tag, self._sbml_tag('listOfReactions'))
        for reaction in model.reactions:
            reaction_tag = ET.SubElement(reactions, self._sbml_tag('reaction'))
            reaction_tag.set(self._sbml_tag('id'), next(reaction_id))
            reaction_tag.set(self._sbml_tag('name'), reaction)
            reaction_tag.set(
                self._sbml_tag('reversible'),
                'true' if model.is_reversible(reaction) else 'false')

            reactants = ET.SubElement(
                reaction_tag, self._sbml_tag('listOfReactants'))
            products = ET.SubElement(
                reaction_tag, self._sbml_tag('listOfProducts'))

            for compound, value in model.get_reaction_values(reaction):
                dest_list = reactants if value < 0 else products
                spec_ref = ET.SubElement(
                    dest_list, self._sbml_tag('speciesReference'))
                spec_ref.set(
                    self._sbml_tag('species'), model_species[compound])
                spec_ref.set(self._sbml_tag('stoichiometry'), str(abs(value)))

            # Create COBRA-compliant parameter list
            kl_tag = ET.SubElement(reaction_tag, self._sbml_tag('kineticLaw'))
            math_tag = ET.SubElement(kl_tag, self._sbml_tag('math'))
            ci_tag = ET.SubElement(math_tag, _tag('ci', MATHML_NS))
            ci_tag.text = 'FLUX_VALUE'
            param_list = ET.SubElement(
                kl_tag, self._sbml_tag('listOfParameters'))
            ET.SubElement(param_list, self._sbml_tag('parameter'), {
                self._sbml_tag('id'): 'LOWER_BOUND',
                self._sbml_tag('name'): 'LOWER_BOUND',
                self._sbml_tag('value'): str(model.limits[reaction].lower)
            })
            ET.SubElement(param_list, self._sbml_tag('parameter'), {
                self._sbml_tag('id'): 'UPPER_BOUND',
                self._sbml_tag('name'): 'UPPER_BOUND',
                self._sbml_tag('value'): str(model.limits[reaction].upper)
            })

        tree = ET.ElementTree(root)
        tree.write(file, default_namespace=self._namespace)



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.datasource import native
from psamm.reaction import Reaction, Compound

from six import StringIO


class TestYAMLDataSource(unittest.TestCase):
    def test_parse_reaction_list(self):
        reactions = list(native.parse_reaction_list('./test.yaml', [
            {
                'id': 'rxn1',
                'equation': {
                    'reversible': True,
                    'left': [
                        { 'id': 'A', 'value': 1 },
                        { 'id': 'B', 'value': 2 } ],
                    'right': [
                        { 'id': 'C', 'value': 1 }
                    ]
                }
            }
        ]))

        self.assertEqual(len(reactions), 1)

        reaction = Reaction(Reaction.Bidir,
                            [(Compound('A'), 1), (Compound('B'), 2)],
                            [(Compound('C'), 1)])
        self.assertEqual(reactions[0].equation, reaction)

    def test_parse_reaction_list_missing_value(self):
        with self.assertRaises(native.ParseError):
            reactions = list(native.parse_reaction_list('./test.yaml', [
                {
                    'id': 'rxn1',
                    'equation': {
                        'left': [
                            { 'id': 'A' }
                        ]
                    }
                }
            ]))

    def test_parse_medium_table(self):
        table = '''
ac      e
glcD    e       -10
co2     e       -       50
'''

        medium = list(native.parse_medium_table_file(StringIO(table.strip())))
        self.assertEqual(len(medium), 3)
        self.assertEqual(medium[0], (Compound('ac', 'e'), None, None, None))
        self.assertEqual(medium[1], (Compound('glcD', 'e'), None, -10, None))
        self.assertEqual(medium[2], (Compound('co2', 'e'), None, None, 50))

    def test_parse_medium(self):
        medium = list(native.parse_medium({
            'compartment': 'e',
            'compounds': [
                {'id': 'ac'},
                {'id': 'glcD', 'lower': -10},
                {'id': 'co2', 'upper': 50},
                {'id': 'compound_x', 'compartment': 'c'},
                {'id': 'compound_y', 'reaction': 'EX_cpdy'}
            ]
        }))

        self.assertEqual(len(medium), 5)
        self.assertEqual(medium[0], (Compound('ac', 'e'), None, None, None))
        self.assertEqual(medium[1], (Compound('glcD', 'e'), None, -10, None))
        self.assertEqual(medium[2], (Compound('co2', 'e'), None, None, 50))
        self.assertEqual(
            medium[3], (Compound('compound_x', 'c'), None, None, None))
        self.assertEqual(
            medium[4], (Compound('compound_y', 'e'), 'EX_cpdy', None, None))


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015  Keith Dufault-Thompson <keitht547@my.uri.edu>

import random
import logging

from ..command import Command, SolverCommandMixin, CommandError
from .. import fluxanalysis, util

logger = logging.getLogger(__name__)


class RandomSparseNetworkCommand(SolverCommandMixin, Command):
    """Find a random minimal network of model reactions.

    Given a reaction to optimize and a threshold, delete reactions randomly
    until the flux of the reaction to optimize falls under the threshold.
    Keep deleting reactions until no more reactions can be deleted. By default
    this uses standard FBA (not tFBA). Since the internal fluxes are irrelevant
    the FBA and tFBA are equivalent for this purpose.

    The threshold can be specified as an absolute flux (e.g. '1.23') or a
    relative flux of the full model flux (e.g. '40.5%').
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--tfba', help='Enable thermodynamic constraints on FBA',
            action='store_true', default=False)
        parser.add_argument('--objective', help='Reaction flux to maximize')
        parser.add_argument(
            'threshold', help='Threshold of max reaction flux',
            type=util.MaybeRelative)
        parser.add_argument(
            '--exchange', help='Only analyze the exchange reaction in model',
            action='store_true')
        super(RandomSparseNetworkCommand, cls).init_parser(parser)

    def run(self):
        if self._args.objective is not None:
            reaction = self._args.objective
        else:
            reaction = self._model.get_biomass_reaction()
            if reaction is None:
                raise CommandError('The biomass reaction was not specified')

        if not self._mm.has_reaction(reaction):
            raise CommandError('Specified reaction is not in model: {}'.format(
                reaction))

        if not self._args.tfba:
            solver = self._get_solver()
        else:
            solver = self._get_solver(integer=True)

        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)

        if self._args.tfba:
            p.add_thermodynamic()

        threshold = self._args.threshold
        if threshold.relative:
            p.maximize(reaction)
            threshold.reference = p.get_flux(reaction)

        flux_threshold = float(threshold)

        logger.info('Flux threshold for {} is {}'.format(
            reaction, flux_threshold))

        essential = {reaction}
        deleted = set()
        if self._args.exchange:
            reactions = set()
            for reaction_id in self._mm.reactions:
                if self._mm.is_exchange(reaction_id):
                    reactions.add(reaction_id)
        else:
            reactions = set(self._mm.reactions)

        test_set = reactions - essential

        while len(test_set) > 0:
            testing_reaction = random.sample(test_set, 1)[0]
            test_set.remove(testing_reaction)

            flux_var = p.get_flux_var(testing_reaction)
            c, = p.prob.add_linear_constraints(flux_var == 0)

            logger.info('Trying FBA without reaction {}...'.format(
                testing_reaction))

            try:
                p.maximize(reaction)
            except fluxanalysis.FluxBalanceError:
                logger.info(
                    'FBA is infeasible, marking {} as essential'.format(
                        testing_reaction))
                c.delete()
                essential.add(testing_reaction)
                continue

            logger.debug('Reaction {} has flux {}'.format(
                reaction, p.get_flux(reaction)))

            if p.get_flux(reaction) < flux_threshold:
                c.delete()
                essential.add(testing_reaction)
                logger.info('Reaction {} was essential'.format(
                    testing_reaction))
            else:
                deleted.add(testing_reaction)
                logger.info('Reaction {} was deleted'.format(testing_reaction))

        for reaction_id in sorted(reactions):
            value = 0 if reaction_id in deleted else 1
            print('{}\t{}'.format(reaction_id, value))



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.datasource import kegg
from psamm.reaction import Reaction, Compound
from psamm.expression.affine import Expression


class TestKEGG(unittest.TestCase):
    def test_kegg_parse(self):
        r = kegg.parse_reaction('C00013 + C00001 <=> 2 C00009')
        self.assertEqual(r, Reaction(Reaction.Bidir, [(Compound('C00013'), 1), (Compound('C00001'), 1)],
                                      [(Compound('C00009'), 2)]))

    def test_kegg_parse_with_count_expression(self):
        r = kegg.parse_reaction('2n C00404 + n C00001 <=> (n+1) C02174')
        self.assertEqual(r, Reaction(Reaction.Bidir, [(Compound('C00404'), Expression('2n')),
                                                       (Compound('C00001'), Expression('n'))],
                                      [(Compound('C02174'), Expression('n+1'))]))

    def test_kegg_parse_with_compound_argument(self):
        r = kegg.parse_reaction('C00039(n) <=> C00013 + C00039(n+1)')
        self.assertEqual(r, Reaction(Reaction.Bidir, [(Compound('C00039', arguments=[Expression('n')]), 1)],
                                      [(Compound('C00013'), 1), (Compound('C00039', arguments=[Expression('n+1')]), 1)]))


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import sys

from ..datasource import sbml
from ..command import Command


class SBMLExport(Command):
    """Export model as SBML file."""

    def run(self):
        writer = sbml.SBMLWriter()
        writer.write_model(
            sys.stdout, self._mm, self._model.parse_compounds())



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.expression.boolean import Expression, Variable, And, Or


class TestVariable(unittest.TestCase):
    def test_variable_init(self):
        v = Variable('x')

    def test_variable_init_long_symbol(self):
        v = Variable('xyz')

    def test_variable_init_number_symbol(self):
        v = Variable('x2')

    def test_variable_init_underscore_symbol(self):
        v = Variable('x_y')

    def test_variable_symbol(self):
        self.assertEqual(Variable('x').symbol, 'x')

    def test_variable_substitute(self):
        subst = {'x': True}
        self.assertEqual(Variable('x').substitute(lambda v: subst.get(v.symbol, v)), True)

    def test_variable_substitute_self(self):
        self.assertEqual(Variable('x').substitute(lambda v: v), Variable('x'))

    def test_variable_substitute_multiple(self):
        subst = {'y': True, 'x': False}
        self.assertEqual(Variable('x').substitute(lambda v: subst.get(v.symbol, v)), False)

    def test_variables_equals_true(self):
        self.assertEqual(Variable('x'), Variable('x'))

    def test_variables_not_equals(self):
        self.assertNotEqual(Variable('x'), Variable('y'))

    def test_variable_not_equals_bool(self):
        self.assertNotEqual(Variable('x'), True)

    def test_variable_hash(self):
        self.assertEqual(hash(Variable('xyz')), hash(Variable('xyz')))

class TestExpression(unittest.TestCase):
    def test_expression_substitute_existing(self):
        e = Expression('b1')
        self.assertTrue(e.substitute(lambda v: {'b1': True}.get(v.symbol, v)))

    def test_expression_substitute_existing_false(self):
        e = Expression('b1')
        self.assertFalse(e.substitute(lambda v: {'b1': False}.get(v.symbol, v)))

    def test_expression_substitute_unknown_to_variable(self):
        e = Expression('b1')
        self.assertEqual(e.substitute(lambda v: v), Variable('b1'))

    def test_expression_substitute_unknown_to_expression(self):
        e = Expression('b1 and b2')
        self.assertEqual(e.substitute(lambda v: v), Expression('b1 and b2'))

    def test_expression_parse_and(self):
        e = Expression('b1 and b2')
        self.assertEqual(e, And(Variable('b1'), Variable('b2')))

    def test_expression_parse_or(self):
        e = Expression('b1 or b2')
        self.assertEqual(e, Or(Variable('b1'), Variable('b2')))

    def test_expression_parse_multiple_with_duplicates(self):
        e = Expression('b1 and b2 and b1 and b4')
        self.assertEqual(e, And(Variable('b1'), Variable('b2'), Variable('b4')))

    def test_expression_parse_multiple_and(self):
        e = Expression('b1 and b2 and b3 and b4')
        self.assertEqual(e, And(Variable('b1'), Variable('b2'), Variable('b3'), Variable('b4')))

    def test_expression_parse_multiple_or(self):
        e = Expression('b1 or b2 or b3 or b4')
        self.assertEqual(e, Or(Variable('b1'), Variable('b2'), Variable('b3'), Variable('b4')))

    def test_expression_parse_multiple_parenthesis_and(self):
        e = Expression('b1 and (b2 and b3) and b4')
        self.assertEqual(e, And(Variable('b1'), Variable('b2'), Variable('b3'), Variable('b4')))

    def test_expression_parse_multiple_parenthesis_or(self):
        e = Expression('b1 or (b2 or b3) or b4')
        self.assertEqual(e, Or(Variable('b1'), Variable('b2'), Variable('b3'), Variable('b4')))

    def test_expression_parse_parentheses_mixed_1(self):
        e = Expression('(b1 and b2) or (b3 and b4)')
        self.assertEqual(e, Or(And(Variable('b1'), Variable('b2')), And(Variable('b3'), Variable('b4'))))

    def test_expression_parse_parentheses_mixed_2(self):
        e = Expression('(b1 or b2) and (b3 or b4)')
        self.assertEqual(e, And(Or(Variable('b1'), Variable('b2')), Or(Variable('b3'), Variable('b4'))))

    def test_expression_parse_parenthesis_with_space_right(self):
        e = Expression('(b1 and b2 )')
        self.assertEqual(e, And(Variable('b1'), Variable('b2')))

    def test_expression_parse_parenthesis_with_space_left(self):
        e = Expression('( b1 and b2)')
        self.assertEqual(e, And(Variable('b1'), Variable('b2')))

    def test_expression_parse_parentheses_right_nested(self):
        e = Expression('(b1 or (b2 and (b3 or (b4 and (b5)))))')
        self.assertEqual(e, Or(Variable('b1'), And(Variable('b2'), Or(Variable('b3'), And(Variable('b4'), Variable('b5'))))))

    def test_expression_parse_parentheses_left_nested(self):
        e = Expression('(((b1 or b2) and b3) or b4) and (b5)')
        self.assertEqual(e, And(Variable('b5'), Or(Variable('b4'), And(Variable('b3'), Or(Variable('b1'), Variable('b2'))))))

    def test_expression_parse_implicit_mixed_1(self):
        e = Expression('b1 and b2 or b3 and b4')
        self.assertEqual(e, Or(And(Variable('b1'), Variable('b2')), And(Variable('b3'), Variable('b4'))))

    def test_expression_parse_implicit_mixed_2(self):
        e = Expression('b1 or b2 and b3 or b4')
        self.assertEqual(e, Or(Variable('b1'), And(Variable('b2'), Variable('b3')), Variable('b4')))

    def test_expression_parse_longer_names(self):
        e = Expression('b12345 and bxyz and testtesttest')
        self.assertEqual(e, And(Variable('b12345'), Variable('bxyz'), Variable('testtesttest')))

    def test_expression_parse_name_starting_with_or(self):
        e = Expression('b1 and order')
        self.assertEqual(e, And(Variable('b1'), Variable('order')))

    def test_expression_parse_name_starting_with_and(self):
        e = Expression('b1 or anderson')
        self.assertEqual(e, Or(Variable('b1'), Variable('anderson')))

    def test_expression_parse_with_extra_space(self):
        e = Expression('b1    and   b2       and b3')
        self.assertEqual(e, And(Variable('b1'), Variable('b2'), Variable('b3')))


if __name__ == '__main__':
    unittest.main()
